/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

/*
 * Test case: 6*n nodes on a six lane highway
 * 
 * 6*n nodes are put on a highway with 6 lanes. Each lane is 5 meters apart
 * from neighboring lanes. Cars are spaced at 90 meters per lane (15
 * meters along the x axis) yielding a total of 66.6 nodes per kilometer.
 */

#include "ns3/core-module.h"
#include "ns3/common-module.h"
#include "ns3/node-module.h"
#include "ns3/helper-module.h"
#include "ns3/mobility-module.h"
#include "ns3/contrib-module.h"
#include "ns3/wifi-module.h"
#include "ns3/traffic-application.h"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <map>
#include <numeric>
#include <sys/time.h>
#include <sys/times.h>

NS_LOG_COMPONENT_DEFINE ("Main");

using namespace ns3;

#include "eventtrace.h"

// very simple pseudo-random number generator for determinstic times
double simpleRandom(unsigned int i)
{
  return ((i * 1549321 + 56273) % 10000) / 10000.0;
}

// time is measured using gettimeofday()
inline double timestamp()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec + tv.tv_usec * 0.000001;
}

class Experiment
{
public:

  static const double simulationTime = 60.0;

  int           m_numNodes;
  int           m_measureNode;

  NodeContainer nodes;

  unsigned int  transmitted;
  unsigned int  received;
  unsigned int  collisions;

  typedef std::map<Address, uint32_t> addressmap_type;
  std::map<Address, uint32_t> addressmap;

  void
  AppTxTrace (std::string context, Ptr<const Packet> p)
  {
    //std::cerr << context << " TX to=" << address << " p: " << p->GetSize() << std::endl;
    ++transmitted;
  }

  void
  AppRxTrace (std::string context, Ptr<const Packet> p, const Address& from)
  {
    //std::cerr << context << " RX from=" << from << " p=" << p->GetSize() << std::endl;
    ++received;
  }

  void
  DevTxTrace (std::string context, Ptr<const Packet> p, Mac48Address address)
  {
    //std::cerr << context << " TX to=" << address << " p: " << p->GetSize() << std::endl;
    // ++transmitted;
  }

  void
  DevRxTrace (std::string context, Ptr<const Packet> p, Mac48Address address)
  {
    // std::cerr << context << " RX to=" << address << " p: " << p->GetSize() << std::endl;
  }

  void
  PhyTxTrace (std::string context, Ptr<const Packet> packet, Ptr<const WifiPhyTag> phytag)
  {
    std::cerr << context << " PHYTX p: " << packet << std::endl;
  }

  void
  PhyRxOkTrace (std::string context, Ptr<const Packet> packet, Ptr<const WifiPhyTag> phytag)
  {
    // std::cerr << context << " PHYRXOK p: " << packet << std::endl;
  }

  void
  PhyRxErrorTrace (std::string context, Ptr<const Packet> packet, Ptr<const WifiPhyTag> phytag, WifiPhy::RxErrorReason reason)
  {
    // if (reason == WifiPhy::RXERROR_LOW_SIGNAL) return;

    // std::cerr << "PHYRXERROR reason=" << WifiPhy::RxErrorReasonToString(reason) << " phytag={" << *phytag << "} " << *packet << std::endl;
    ++collisions;
  }

  void
  Run (int numNodes)
  {
    m_numNodes = numNodes;

    // Create nodes

    nodes.Create (numNodes);
  
    // Install wifi devices on the nodes

    Ns2ExtWifiChannelHelper wifiChannel = Ns2ExtWifiChannelHelper::Default ();
    wifiChannel.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");
    wifiChannel.AddPropagationLoss("ns3::ThreeLogDistancePropagationLossModel");

    Ns2ExtWifiPhyHelper wifiPhy = Ns2ExtWifiPhyHelper::Default();
    wifiPhy.SetChannel(wifiChannel.Create());
    wifiPhy.Set("UseConstantNoiseFloor", BooleanValue(true));
    wifiPhy.Set("ConstantNoiseFloor", DoubleValue(-99.0));
    wifiPhy.Set("PreambleCapture", BooleanValue(true));
    wifiPhy.Set("DataCapture", BooleanValue(true));

    NqosWifiMacHelper wifiMac = NqosWifiMacHelper::Default ();
    wifiMac.SetType ("ns3::AdhocWifiMac");

    WifiHelper wifi = WifiHelper::Default();
    wifi.SetStandard (WIFI_PHY_STANDARD_80211a);
    wifi.SetRemoteStationManager("ns3::ConstantRateWifiManager",
                                 "DataMode", StringValue("wifia-6mbs"),
                                 "NonUnicastMode", StringValue("wifia-6mbs"));

    wifi.Install(wifiPhy, wifiMac, nodes);

    // set nodes on to highway lanes

    MobilityHelper mobility;
    Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
    for (int i = 0; i < numNodes; ++i)
      {
        positionAlloc->Add (Vector (i * 15, (i % 6) * 5, 0.0));
      }
    mobility.SetPositionAllocator (positionAlloc);
    mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
    mobility.Install (nodes);

    // add packet socket handlers

    PacketSocketHelper packetSocket;
    packetSocket.Install (nodes);

    // create mapping of (mac) addresses to node ids

    {
      addressmap.clear();

      for (NodeContainer::Iterator ci = nodes.Begin(); ci != nodes.End(); ++ci)
        {
          // std::cerr << "Id: " << (*ci)->GetId() << "\n";
          // std::cerr << "Dev: " << (*ci)->GetDevice(0)->GetAddress() << "\n";
          addressmap[ (*ci)->GetDevice(0)->GetAddress() ] = (*ci)->GetId();
        }
    }

    // Configure packet socket: use broadcasts


    PacketSocketAddress socketBroadcast;
    socketBroadcast.SetAllDevices();
    socketBroadcast.SetPhysicalAddress(Mac48Address::GetBroadcast());
    socketBroadcast.SetProtocol(1);

    // Install TrafficApplications on Nodes

    Ptr<SimpleTrafficPacketFactory> packetFactory = CreateObjectWithAttributes<SimpleTrafficPacketFactory>("Size", UintegerValue(400));

    TrafficHelper trafficApp("ns3::PacketSocketFactory", socketBroadcast);
    trafficApp.SetAttribute("PacketFactory", PointerValue(packetFactory) );
    trafficApp.SetAttribute("OnTime", RandomVariableValue( ConstantVariable(simulationTime) ));
    trafficApp.SetAttribute("OffTime", RandomVariableValue( UniformVariable(0.0, 1.0 / 10.0) ));
    //trafficApp.SetAttribute("OffTime", RandomVariableValue( ConstantVariable(0.0) ));
    trafficApp.SetAttribute("Interval", RandomVariableValue( ConstantVariable(1.0 / 10.0) ));

    ApplicationContainer app = trafficApp.Install(nodes);
    for (unsigned int ai = 0; ai < app.GetN(); ++ai)
      {
        // change OffTimes to deterministic values
        app.Get(ai)->SetAttribute ("OffTime", RandomVariableValue (ConstantVariable ( simpleRandom(ai) * 0.1 )));
        // std::cerr << "Node " << ai << " offset " << std::setprecision(17) << (simpleRandom(ai) * 0.1)  << "\n";
      }
    app.Start (Seconds (0.0));
    app.Stop (Seconds (simulationTime));

    // Install PacketSinks on listener node

    PacketSinkHelper sink("ns3::PacketSocketFactory", socketBroadcast);

    ApplicationContainer appSinks = sink.Install(nodes);
    appSinks.Start(Seconds(0.0));
    appSinks.Stop(Seconds(simulationTime+1.0));

    // Add Trace Callbacks

    Config::Connect ("/NodeList/*/ApplicationList/*/$ns3::TrafficApplication/Tx", MakeCallback (&Experiment::AppTxTrace, this));
    Config::Connect ("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback (&Experiment::AppRxTrace, this));

    //m_measureNode = m_numNodes / 2;

    //std::ostringstream measureRxTrace;
    //measureRxTrace << "/NodeList/" << m_measureNode << "/ApplicationList/*/$ns3::TrafficApplication/Rx";
    //Config::Connect (measureRxTrace.str(), MakeCallback (&Experiment::AppRxTrace, this));

    // Config::Connect ("/NodeList/*/DeviceList/*/Tx", MakeCallback (&Experiment::DevTxTrace, this));

    // Config::Connect ("/NodeList/2/DeviceList/*/Tx", MakeCallback (&Experiment::DevTxTrace2, this));
    // Config::Connect ("/NodeList/*/DeviceList/*/Phy/State/Tx", MakeCallback (&Experiment::PhyTxTrace, this));

    // Config::Connect ("/NodeList/*/DeviceList/*/Phy/State/RxOk", MakeCallback (&Experiment::PhyRxOkTrace, this));
    //Config::Connect ("/NodeList/*/DeviceList/*/Phy/State/RxError", MakeCallback (&Experiment::PhyRxErrorTrace, this));

    // ConnectEventTrace();

    // Wifi Traces

    // std::ofstream ascii;
    // ascii.open ("wifi-test.tr");
    // trace = &ascii;
    //Ns2ExtWifiPhyHelper::EnableAsciiAll (std::cout);
    // WifiHelper::EnablePcapAll ("wifi-test");

    transmitted = 0;
    received = 0;
    collisions = 0;

    Simulator::Run ();

    Simulator::Destroy ();
  }
};

template <typename Container>
double calcMeanValue(const Container& c)
{
  return std::accumulate(c.begin(), c.end(), 0.0) / c.size();
}

template <typename Container>
double calcStandardDeviation(const Container& c)
{
  double squareSum = 0.0;
  double sum = 0.0;

  for (typename Container::const_iterator ei = c.begin();
       ei != c.end(); ++ei)
    {
      squareSum += (*ei) * (*ei);
      sum += *ei;
    }
  
  double mean = sum / c.size();
  return sqrt( (squareSum / c.size()) - (mean * mean) );
}

template <typename Container>
double calcConfidenceInterval(const Container& c)
{
  return 2.576 * calcStandardDeviation(c) / sqrt(c.size());
}

int main (int argc, char *argv[])
{
  // disable fragmentation
  Config::SetDefault ("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue ("2200"));
  Config::SetDefault ("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue ("2200"));

  CommandLine cmd;

  int replications = 1;
  int fixedNumNodes = 0;

  cmd.AddValue("NumNodes", "Perform run for fixed number of nodes", fixedNumNodes);
  cmd.AddValue("Replications", "Perform independent replications", replications);
  cmd.Parse (argc, argv);

  double clockticks = sysconf(_SC_CLK_TCK);

  // FriisPropagationLossModel
  for (int numNodes = 6; numNodes <= 180; numNodes += 6)
    {
      if (fixedNumNodes != 0 && fixedNumNodes != numNodes) continue;

      std::vector<double> transmitted;
      std::vector<double> received;
      std::vector<double> tsdelta_list;
      std::vector<double> tudelta_user_list;
      std::vector<double> tudelta_system_list;
      std::vector<double> tcdelta_list;

      for(int rep = 0; rep < replications; ++rep)
        {
          struct tms tu1;
          clock_t tc1 = times(&tu1);
          double ts1 = timestamp();

          SeedManager::SetRun(rep);

          Experiment experiment;
          experiment.Run (numNodes);

          double ts2 = timestamp();
          
          struct tms tu2;
          clock_t tc2 = times(&tu2);

          double tsdelta = ts2 - ts1;
          double tcdelta = (double)(tc2 - tc1) / clockticks;
          double tudelta_user = (double)(tu2.tms_utime - tu1.tms_utime) / clockticks;
          double tudelta_system = (double)(tu2.tms_stime - tu1.tms_stime) / clockticks;
          
          std::cout << "# " << numNodes
                    << " " << experiment.transmitted
                    << " " << experiment.received
                    << " " << tsdelta
                    << " " << tudelta_user
                    << " " << tudelta_system
                    << " " << tcdelta
                    << "\n";

          transmitted.push_back( experiment.transmitted );
          received.push_back( experiment.received );
          tsdelta_list.push_back(tsdelta);
          tudelta_user_list.push_back(tudelta_user);
          tudelta_system_list.push_back(tudelta_system);
          tcdelta_list.push_back(tcdelta);
        }

      std::cout << std::fixed
                << numNodes
                << " " << calcMeanValue(transmitted) << " " << calcConfidenceInterval(transmitted)
                << " " << calcMeanValue(received) << " " << calcConfidenceInterval(received)
                << " " << calcMeanValue(tsdelta_list) << " " << calcConfidenceInterval(tsdelta_list)
                << " " << calcMeanValue(tudelta_user_list) << " " << calcConfidenceInterval(tudelta_user_list)
                << " " << calcMeanValue(tudelta_system_list) << " " << calcConfidenceInterval(tudelta_system_list)
                << " " << calcMeanValue(tcdelta_list) << " " << calcConfidenceInterval(tcdelta_list)
                << std::endl;
    }

  return 0;
}
