/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Timo Bingmann <timo.bingmann@student.kit.edu>
 */

/*
 * Test case: Three node hidden terminal interference
 * 
 * Three nodes 0, 1 and 2 are create along the x axis with 0 at the origin. 1
 * and 2 send packets to 0. The distance 2 - 0 is kept constant while 1's
 * distance to 0 varies with each run.
 */

#include "ns3/core-module.h"
#include "ns3/common-module.h"
#include "ns3/node-module.h"
#include "ns3/helper-module.h"
#include "ns3/mobility-module.h"
#include "ns3/contrib-module.h"
#include "ns3/onoff-application.h"
#include "ns3/wifi-mode.h"
#include "ns3/wifi-preamble.h"
#include "ns3/propagation-loss-model.h"

#include <iostream>
#include <fstream>
#include <map>
#include <set>
#include "assert.h"

NS_LOG_COMPONENT_DEFINE ("Main");

using namespace ns3;

#include "eventtrace.h"

std::ostringstream plot1, plot2, plot3, plot4, plot5;

template <typename Container, typename Element>
bool contains(const Container& container, const Element& element)
{
  return (container.find(element) != container.end());
}

class Experiment
{
public:

  double        m_node1pos;
  double        m_node2pos;
  double        m_timedelta;

  Address       m_packetSource1;
  Address       m_packetSource2;

  std::vector<uint64_t> m_sent1;
  std::vector<uint64_t> m_sent2;

  std::map<uint64_t, Time> m_packetsenttime;

  std::set<uint64_t> m_receivedAt0;
  std::set<uint64_t> m_receivedAt1;
  std::set<uint64_t> m_receivedAt2;

  std::set<uint64_t> m_errorPreambleCapture;
  std::set<uint64_t> m_errorDataCapture;

  void Output()
  {
    int val1 = 0, val2 = 0, val3 = 0, val4 = 0, val5 = 0;
    int valcount = 0;

    for(unsigned int i = 0; i < m_sent1.size() && i < m_sent2.size(); ++i)
      {
        uint64_t packetid1 = m_sent1[i];
        uint64_t packetid2 = m_sent2[i];

        Time time1 = m_packetsenttime[ packetid1 ];
        Time time2 = m_packetsenttime[ packetid2 ];

        double timedelta = (time1 - time2).GetSeconds();
        
        if ( fabs(timedelta - m_timedelta) > 1e-9)
          {
            // Send time were further apart than expected: CSMA prohibited collision.
            val1++;
          }
        else
          {
            if ( contains(m_receivedAt0, packetid1) )
              {
                val5++;

                if ( contains(m_errorDataCapture, packetid2) )
                  val4++;
                else if ( contains(m_errorPreambleCapture, packetid2) )
                  val3++;
                else
                  val2++;
              }

            if ( contains(m_receivedAt0, packetid2) )
              {
                // val2 = 2;
              }
          }
        valcount++;
      }

    plot1 << m_node1pos << " " << m_timedelta << " " << (val1 / (double)valcount) << "\n";
    plot2 << m_node1pos << " " << m_timedelta << " " << (val2 / (double)valcount) << "\n";
    plot3 << m_node1pos << " " << m_timedelta << " " << (val3 / (double)valcount) << "\n";
    plot4 << m_node1pos << " " << m_timedelta << " " << (val4 / (double)valcount) << "\n";
    plot5 << m_node1pos << " " << m_timedelta << " " << (val5 / (double)valcount) << "\n";

    std::cerr << m_node1pos << " " << m_timedelta << " " << (val1 / (double)valcount) << "\n";
    std::cerr << m_node1pos << " " << m_timedelta << " " << (val2 / (double)valcount) << "\n";
    std::cerr << m_node1pos << " " << m_timedelta << " " << (val3 / (double)valcount) << "\n";
    std::cerr << m_node1pos << " " << m_timedelta << " " << (val4 / (double)valcount) << "\n";
    std::cerr << m_node1pos << " " << m_timedelta << " " << (val5 / (double)valcount) << "\n";
  }

private:

  void
  PacketSinkRxTrace0(std::string context, Ptr<const Packet> p, const Address &ad)
  {
    m_receivedAt0.insert(p->GetUid());
  }

  void
  PacketSinkRxTrace1(std::string context, Ptr<const Packet> p, const Address &ad)
  {
    m_receivedAt1.insert(p->GetUid());
  }

  void
  PacketSinkRxTrace2(std::string context, Ptr<const Packet> p, const Address &ad)
  {
    m_receivedAt2.insert(p->GetUid());
  }

  void
  DevTxTrace1 (std::string context, Ptr<const Packet> p, Mac48Address address)
  {
    // std::cout << context << " " << Simulator::Now() << " TX to=" << address << " p: " << p->GetSize() << std::endl;

    m_sent1.push_back(p->GetUid());
    m_packetsenttime[p->GetUid()] = Simulator::Now();
  }

  void
  DevTxTrace2 (std::string context, Ptr<const Packet> p, Mac48Address address)
  {
    // std::cout << context << " " << Simulator::Now() << " TX to=" << address << " p: " << p->GetSize() << std::endl;

    m_sent2.push_back(p->GetUid());
    m_packetsenttime[p->GetUid()] = Simulator::Now();
  }

  void
  DevPhyTxTrace1 (std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> tag)
  {
    // std::cout << context << " " << Simulator::Now() << " TX p: " << p->GetSize() << std::endl;

    m_sent1.push_back(p->GetUid());
    m_packetsenttime[p->GetUid()] = Simulator::Now();
  }

  void
  DevPhyTxTrace2 (std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> tag)
  {
    // std::cout << context << " " << Simulator::Now() << " TX p: " << p->GetSize() << std::endl;

    m_sent2.push_back(p->GetUid());
    m_packetsenttime[p->GetUid()] = Simulator::Now();
  }

  void
  DevPhyRxErrorTrace0 (std::string context, Ptr<const Packet> p, Ptr<const WifiPhyTag> phytag, WifiPhy::RxErrorReason rxErrorReason)
  {
    // std::cerr << context << " " << Simulator::Now() << " PHYRXERROR" << " reason=" << WifiPhy::RxErrorReasonToString(rxErrorReason) << " phytag={" << *phytag << "} packet={" << *p << "}" << std::endl;

    if (rxErrorReason == WifiPhy::RXERROR_CAPTURE_PREAMBLE)
      m_errorPreambleCapture.insert( p->GetUid() );

    if (rxErrorReason == WifiPhy::RXERROR_CAPTURE_DATA)
      m_errorDataCapture.insert( p->GetUid() );
  }

public:
  void
  Run (const WifiHelper &wifi, const WifiPhyHelper &wifiPhy, const WifiMacHelper& wifiMac, double node1pos, double node2pos, double timedelta)
  {
    const double simulationTime = 100.0;

    m_node1pos = node1pos;
    m_node2pos = node2pos;
    m_timedelta = timedelta;

    NodeContainer nodes;
    nodes.Create (3);

    // Install WifiPhy and set up mobility
    wifi.Install (wifiPhy, wifiMac, nodes);

    MobilityHelper mobility;
    Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator> ();
    positionAlloc->Add (Vector (0.0, 0.0, 0.0));
    positionAlloc->Add (Vector (node1pos, 0.0, 0.0));
    positionAlloc->Add (Vector (node2pos, 0.0, 0.0));
    mobility.SetPositionAllocator (positionAlloc);
    mobility.SetMobilityModel ("ns3::ConstantPositionMobilityModel");
    mobility.Install (nodes);

    // Add packet socket handlers

    PacketSocketHelper packetSocket;
    packetSocket.Install (nodes);

    // Configure packet socket: use broadcasts

    PacketSocketAddress socket;
    socket.SetAllDevices();
    socket.SetPhysicalAddress(Mac48Address::GetBroadcast());
    socket.SetProtocol(1);

    // Install PacketSink on all nodes

    PacketSinkHelper sink("ns3::PacketSocketFactory", Address(socket));

    ApplicationContainer app = sink.Install(nodes);
    app.Start(Seconds(0.0));
    app.Stop(Seconds(simulationTime));

    Config::Connect("/NodeList/0/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&Experiment::PacketSinkRxTrace0, this));
    Config::Connect("/NodeList/1/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&Experiment::PacketSinkRxTrace1, this));
    Config::Connect("/NodeList/2/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&Experiment::PacketSinkRxTrace2, this));

    // Save packet source addresses

    m_packetSource1 = nodes.Get(1)->GetDevice(0)->GetAddress();
    m_packetSource2 = nodes.Get(2)->GetDevice(0)->GetAddress();

    // Install OnOffApplication on Node 1

    OnOffHelper onoff1("ns3::PacketSocketFactory", Address(socket));
    onoff1.SetAttribute ("OnTime", RandomVariableValue (ConstantVariable (simulationTime)));
    onoff1.SetAttribute ("OffTime", RandomVariableValue (ConstantVariable (timedelta)));
    onoff1.SetAttribute ("DataRate", DataRateValue (DataRate ("40000B/s")));
    onoff1.SetAttribute ("PacketSize", UintegerValue (200));
    // 40000B/s with 200 B packets equals 200 packet/s or 0.005 intervals

    app = onoff1.Install(nodes.Get(1));
    app.Start(Seconds(0.0));
    app.Stop(Seconds(simulationTime));

    // Install OnOffApplication on Node 2

    OnOffHelper onoff2("ns3::PacketSocketFactory", Address(socket));
    onoff2.SetAttribute ("OnTime", RandomVariableValue (ConstantVariable (simulationTime)));
    onoff2.SetAttribute ("OffTime", RandomVariableValue (ConstantVariable (0)));
    onoff2.SetAttribute ("DataRate", DataRateValue (DataRate ("40000B/s")));
    onoff2.SetAttribute ("PacketSize", UintegerValue (200));
    // 40000B/s with 200 B packets equals 200 packet/s or 0.005 intervals

    app = onoff2.Install(nodes.Get(2));
    app.Start(Seconds(0.0));
    app.Stop(Seconds(simulationTime));

    // Add Trace Callbacks

    Config::Connect ("/NodeList/1/DeviceList/*/Phy/State/Tx", MakeCallback (&Experiment::DevPhyTxTrace1, this));
    Config::Connect ("/NodeList/2/DeviceList/*/Phy/State/Tx", MakeCallback (&Experiment::DevPhyTxTrace2, this));

    Config::Connect ("/NodeList/0/DeviceList/*/Phy/State/RxError", MakeCallback (&Experiment::DevPhyRxErrorTrace0, this));

    //ConnectEventTrace();

    // Wifi Traces

    // std::ofstream ascii;
    // ascii.open ("wifi-test.tr");
    // trace = &ascii;
    // WifiHelper::EnableAsciiAll (ascii);
    // WifiHelper::EnablePcapAll ("wifi-test");

    Simulator::Run ();

    Simulator::Destroy ();
  }
};

int main (int argc, char *argv[])
{
  // disable fragmentation
  Config::SetDefault ("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue ("2200"));
  Config::SetDefault ("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue ("2200"));

  LogComponentEnable("Main", LOG_LEVEL_ALL);

  CommandLine cmd;
  cmd.Parse (argc, argv);

  Ns2ExtWifiChannelHelper wifiChannel;
  // wifiChannel.AddPropagationLoss ("ns3::FriisPropagationLossModel");
  wifiChannel.AddPropagationLoss("ns3::ThreeLogDistancePropagationLossModel");
  wifiChannel.AddPropagationLoss("ns3::NakagamiPropagationLossModel");
  wifiChannel.SetPropagationDelay ("ns3::ConstantSpeedPropagationDelayModel");

  Ns2ExtWifiPhyHelper wifiPhy = Ns2ExtWifiPhyHelper::Default ();
  wifiPhy.Set("UseConstantNoiseFloor", BooleanValue(true));
  wifiPhy.Set("ConstantNoiseFloor", DoubleValue(-99.0));
  wifiPhy.Set("CarrierSenseThreshold", DoubleValue(-82.0));

  NqosWifiMacHelper wifiMac = NqosWifiMacHelper::Default ();
  wifiMac.SetType ("ns3::AdhocWifiMac");

  WifiHelper wifi = WifiHelper::Default ();
  wifi.SetStandard (WIFI_PHY_STANDARD_80211a);
  wifi.SetRemoteStationManager ("ns3::ConstantRateWifiManager",
                                "DataMode", StringValue ("wifia-6mbs"));

  NS_LOG_DEBUG ("Starting");

  double node2pos = -600;

  for (double pos = 0.0; pos <= 1500.0; pos += 20.0)
    {
      for (double timedelta = 0e-6; timedelta <= 500e-6; timedelta += 2e-6)
        {
          std::cerr << "pos " << pos << " timedelta " << timedelta << "\n";

          wifiPhy.SetChannel (wifiChannel.Create ());

          Experiment experiment;
          experiment.Run (wifi, wifiPhy, wifiMac, pos, node2pos, timedelta);
          experiment.Output();
        }
      plot1 << "\n";
      plot2 << "\n";
      plot3 << "\n";
      plot4 << "\n";
      plot5 << "\n";
    }

  std::cout << plot1.str() << "\n\n";
  std::cout << plot2.str() << "\n\n";
  std::cout << plot3.str() << "\n\n";
  std::cout << plot4.str() << "\n\n";
  std::cout << plot5.str() << "\n\n";

  return 0;
}
