/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
//
// Copyright (c) 2009 Timo Bingmann
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Author: Timo Bingmann <timo.bingmann@student.kit.edu>
// Based on onoff-application.cc
//

// ns3 - Traffic Generator Application class

#include "traffic-application.h"

#include "ns3/log.h"
#include "ns3/address.h"
#include "ns3/node.h"
#include "ns3/random-variable.h"
#include "ns3/socket.h"
#include "ns3/simulator.h"
#include "ns3/socket-factory.h"
#include "ns3/packet.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/packet-socket-factory.h"
#include "ns3/uinteger.h"
#include "ns3/pointer.h"

NS_LOG_COMPONENT_DEFINE ("TrafficApplication");

using namespace std;

namespace ns3 {

// ------------------------------------------------------------------------- //

NS_OBJECT_ENSURE_REGISTERED (TrafficApplication);

TypeId
TrafficApplication::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::TrafficApplication")
    .SetParent<Application> ()
    .AddConstructor<TrafficApplication> ()
    .AddAttribute("Interval", "RandomVariable used for stochastic inter-packet intervals (in seconds).",
                  RandomVariableValue (ConstantVariable (1.0)),
                  MakeRandomVariableAccessor (&TrafficApplication::m_interval),
                  MakeRandomVariableChecker ())
    .AddAttribute("OnTime", "A RandomVariable used to pick the duration of the 'On' state.",
                  RandomVariableValue (ConstantVariable (1.0)),
                  MakeRandomVariableAccessor (&TrafficApplication::m_onTime),
                  MakeRandomVariableChecker ())
    .AddAttribute("OffTime", "A RandomVariable used to pick the duration of the 'Off' state.",
                  RandomVariableValue (ConstantVariable (1.0)),
                  MakeRandomVariableAccessor (&TrafficApplication::m_offTime),
                  MakeRandomVariableChecker ())
    .AddAttribute("PacketFactory", "A TrafficPacketFactory object called to create packets to send.",
                  PointerValue ( CreateObject<SimpleTrafficPacketFactory>()),
                  MakePointerAccessor (&TrafficApplication::m_packetFactory),
                  MakePointerChecker<TrafficPacketFactory> ())
    .AddAttribute("Remote", "The address of the destination",
                  AddressValue (),
                  MakeAddressAccessor (&TrafficApplication::m_peer),
                  MakeAddressChecker ())
    .AddAttribute("Protocol", "The type of protocol to use.",
                  TypeIdValue (PacketSocketFactory::GetTypeId ()),
                  MakeTypeIdAccessor (&TrafficApplication::m_protocolTid),
                  MakeTypeIdChecker ())
    .AddTraceSource("Tx", "A new packet is created and is sent",
                    MakeTraceSourceAccessor (&TrafficApplication::m_txTrace))
    .AddTraceSource("Rx", "A packet has been received",
                    MakeTraceSourceAccessor (&TrafficApplication::m_rxTrace))
    ;
  return tid;
}


TrafficApplication::TrafficApplication ()
  : m_socket(NULL),
    m_connected(false),
    m_totalBytes(0),
    m_lastStartTime( Seconds(0) )
{
  NS_LOG_FUNCTION_NOARGS ();
}

TrafficApplication::~TrafficApplication()
{
  NS_LOG_FUNCTION_NOARGS ();
}

uint32_t
TrafficApplication::GetTotalBytes() const
{
  return m_totalBytes;
}

void
TrafficApplication::DoDispose (void)
{
  NS_LOG_FUNCTION_NOARGS ();

  m_socket = NULL;
  Application::DoDispose (); // chain up
}

// Application Methods
void TrafficApplication::StartApplication() // Called at time specified by Start
{
  NS_LOG_FUNCTION_NOARGS ();

  // Create the socket if not already
  if (!m_socket)
    {
      m_socket = Socket::CreateSocket (GetNode(), m_protocolTid);
      m_socket->Bind ();
      m_socket->Connect (m_peer);
    }

  m_socket->SetRecvCallback (MakeCallback(&TrafficApplication::HandleRead, this));

  // Insure no pending event
  CancelEvents ();

  // If we are not yet connected, there is nothing to do here
  // The ConnectionComplete upcall will start timers at that time
  //if (!m_connected) return;
  ScheduleStartEvent();
}

void TrafficApplication::StopApplication() // Called at time specified by Stop
{
  NS_LOG_FUNCTION_NOARGS ();

  CancelEvents ();
  if(m_socket != 0)
    {
      m_socket->Close ();
      m_socket->SetRecvCallback (MakeNullCallback<void, Ptr<Socket> > ());
    }
  else
    {
      NS_LOG_WARN("TrafficApplication found NULL socket to close in StopApplication");
    }
}

void TrafficApplication::CancelEvents ()
{
  NS_LOG_FUNCTION_NOARGS ();

  Simulator::Cancel(m_sendEvent);
  Simulator::Cancel(m_startStopEvent);
}

// Event handlers
void TrafficApplication::StartSending()
{
  NS_LOG_FUNCTION_NOARGS ();

  SendPacket();
  ScheduleStopEvent();
}

void TrafficApplication::StopSending()
{
  NS_LOG_FUNCTION_NOARGS ();

  CancelEvents();
  ScheduleStartEvent();
}

void TrafficApplication::ScheduleStartEvent()
{
  // Schedules the event to start sending data (switch to the "On" state)
  NS_LOG_FUNCTION_NOARGS ();

  Time offInterval = Seconds(m_offTime.GetValue());
  NS_LOG_LOGIC ("start at " << offInterval);
  m_startStopEvent = Simulator::Schedule(offInterval, &TrafficApplication::StartSending, this);
}

void TrafficApplication::ScheduleStopEvent()
{
  // Schedules the event to stop sending data (switch to "Off" state)
  NS_LOG_FUNCTION_NOARGS ();

  Time onInterval = Seconds(m_onTime.GetValue());
  NS_LOG_LOGIC ("stop at " << onInterval);
  m_startStopEvent = Simulator::Schedule(onInterval, &TrafficApplication::StopSending, this);
}

// Private helpers
void TrafficApplication::SendPacket()
{
  NS_LOG_FUNCTION_NOARGS ();

  NS_LOG_LOGIC ("sending packet at " << Simulator::Now());
  NS_ASSERT (m_sendEvent.IsExpired ());
  NS_ASSERT (m_packetFactory != NULL);

  Ptr<Packet> packet = m_packetFactory->CreatePacket();
  m_txTrace (packet);
  m_totalBytes += packet->GetSize();
  m_socket->Send (packet);
  m_lastStartTime = Simulator::Now();
  ScheduleNextTx();
}

void TrafficApplication::ScheduleNextTx()
{
  NS_LOG_FUNCTION_NOARGS ();

  // Time till next packet
  Time nextTime = Seconds ( m_interval.GetValue() );
  NS_LOG_LOGIC ("nextTime = " << nextTime);

  m_sendEvent = Simulator::Schedule(nextTime, 
                                    &TrafficApplication::SendPacket, this);
}

void TrafficApplication::HandleRead (Ptr<Socket> socket)
{
  NS_LOG_FUNCTION (this << socket);
  Ptr<Packet> packet;
  Address from;
  while (packet = socket->RecvFrom (from))
    {
      if (packet->GetSize() == 0) // EOF
        {
	  break;
        }

      m_rxTrace (packet, from);
    }
}
  
void TrafficApplication::ConnectionSucceeded(Ptr<Socket>)
{
  NS_LOG_FUNCTION_NOARGS ();

  m_connected = true;
  ScheduleStartEvent();
}
  
void TrafficApplication::ConnectionFailed(Ptr<Socket>)
{
  NS_LOG_FUNCTION_NOARGS ();
  cout << "TrafficApplication, Connection Failed" << endl;
}

// ------------------------------------------------------------------------- //

NS_OBJECT_ENSURE_REGISTERED(TrafficPacketFactory);

TypeId
TrafficPacketFactory::GetTypeId()
{
  static TypeId tid = TypeId ("ns3::TrafficPacketFactory")
    .SetParent<Object> ();
  return tid;
}

TrafficPacketFactory::TrafficPacketFactory()
{
}

// ------------------------------------------------------------------------- //

NS_OBJECT_ENSURE_REGISTERED (CloneTrafficPacketFactory);

TypeId 
CloneTrafficPacketFactory::GetTypeId()
{
  static TypeId tid = TypeId("ns3::CloneTrafficPacketFactory")
    .SetParent<TrafficPacketFactory>()
/* This does not work because Packet is-NOT-a Object:
    .AddAttribute( "Prototype",
                   "Prototype packet to clone.",
                   PointerValue( Create<Packet>() ),
                   MakePointerAccessor(&ClonePacketFactory::m_prototype),
                   MakePointerChecker<Packet>() )
*/
    ;
  return tid;
}

CloneTrafficPacketFactory::CloneTrafficPacketFactory(Ptr<Packet> prototype)
  : m_prototype(prototype)
{}

void
CloneTrafficPacketFactory::SetPrototype(Ptr<Packet> prototype)
{
  m_prototype = prototype;
}

Ptr<Packet>
CloneTrafficPacketFactory::CreatePacket()
{
  return m_prototype->Copy();
}

// ------------------------------------------------------------------------- //

NS_OBJECT_ENSURE_REGISTERED (SimpleTrafficPacketFactory);

TypeId 
SimpleTrafficPacketFactory::GetTypeId()
{
  static TypeId tid = TypeId("ns3::SimpleTrafficPacketFactory")
    .SetParent<TrafficPacketFactory>()
    .AddAttribute("Size",
                  "Size in bytes of packets generated.",
                  UintegerValue(100),
                  MakeUintegerAccessor(&SimpleTrafficPacketFactory::m_size),
                  MakeUintegerChecker<unsigned int>())
    
    ;
  return tid;
}

SimpleTrafficPacketFactory::SimpleTrafficPacketFactory()
{}

Ptr<Packet>
SimpleTrafficPacketFactory::CreatePacket()
{
  Ptr<Packet> packet = Create<Packet>(m_size);
  return packet;
}

void
SimpleTrafficPacketFactory::SetSize(unsigned int size)
{
  m_size = size;
}

unsigned int
SimpleTrafficPacketFactory::GetSize() const
{
  return m_size;
}

// ------------------------------------------------------------------------- //

} // Namespace ns3
