/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
//
// Copyright (c) 2009 Timo Bingmann
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 2 as
// published by the Free Software Foundation;
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// Author: Timo Bingmann <timo.bingmann@student.kit.edu>
// Based on onoff-application.h
//

// ns3 - Traffic Generator Application class

#ifndef __traffic_application_h__
#define __traffic_application_h__

#include "ns3/address.h"
#include "ns3/application.h"
#include "ns3/event-id.h"
#include "ns3/ptr.h"
#include "ns3/random-variable.h"
#include "ns3/traced-callback.h"

namespace ns3 {

class Socket;
class TrafficPacketFactory;

/**
 * \ingroup applications 
 * \defgroup traffic TrafficApplication
 *
 * This is a traffic generator with stochastic packet sending
 * intervals. Similar to OnOffApplication it follows an On/Off pattern: after
 * Application::StartApplication is called, "On" and "Off" states
 * alternate. The duration of each of these states is determined with the
 * onTime and the offTime random variables. During the "Off" state, no traffic
 * is generated.  During the "On" state, packets are generated by the
 * configured TrafficPacketFactory object. The packet sending intervals are
 * determined by the Interval random variable. The first packet is sent
 * immediately at the beginning of the "On" state, the second after the first
 * time interval pulled from "Interval".
 *
 * Contrary to OnOffApplication, this application also receives incoming
 * packets in the created socket. Thus it contains a subset of PacketSink. It
 * does not accept connections.
 */
 /**
 * \ingroup traffic
 *
 * \brief Generate traffic packets to a single destination with stochastic
 * sending intervals.
 *
 * This is a traffic generator with stochastic packet sending
 * intervals. Similar to OnOffApplication it follows an On/Off pattern: after
 * Application::StartApplication is called, "On" and "Off" states
 * alternate. The duration of each of these states is determined with the
 * onTime and the offTime random variables. During the "Off" state, no traffic
 * is generated.  During the "On" state, packets are generated by the
 * configured TrafficPacketFactory object. The packet sending intervals are
 * determined by the Interval random variable. The first packet is sent
 * immediately at the beginning of the "On" state, the second after the first
 * time interval pulled from "Interval".
 *
 * Contrary to OnOffApplication, this application also receives incoming
 * packets in the created socket. Thus it contains a subset of PacketSink. It
 * does not accept connections.
 */
class TrafficApplication : public Application 
{
public:
  static TypeId GetTypeId ();

  TrafficApplication ();

  virtual ~TrafficApplication();

  /**
   * \brief Return total bytes sent by this object.
   */
  uint32_t      GetTotalBytes() const;

protected:
  virtual void DoDispose ();

private:
  // Inherited from Application base class.
  virtual void StartApplication ();    // Called at time specified by Start
  virtual void StopApplication ();     // Called at time specified by Stop

  // Helpers
  void CancelEvents ();

  // Event handlers
  void StartSending();
  void StopSending();
  void SendPacket();

  Ptr<Socket>     m_socket;       // Associated socket
  TypeId          m_protocolTid;   // protocol type id
  Address         m_peer;         // Peer address
  bool            m_connected;    // True if connected

  RandomVariable  m_onTime;       // rng for On Time
  RandomVariable  m_offTime;      // rng for Off Time
  RandomVariable  m_interval;     // stochastic interval between packets

  Ptr<TrafficPacketFactory> m_packetFactory; // factory called to create packets

  uint32_t        m_totalBytes;   // Total bytes sent so far

  Time            m_lastStartTime; // Time last packet sent
  EventId         m_startStopEvent; // Event id for next start or stop event
  EventId         m_sendEvent;     // Eventid of pending "send packet" event

  /// trace callback for each sent packet
  TracedCallback< Ptr<const Packet> > m_txTrace;

  /// callback for received packets
  TracedCallback< Ptr<const Packet>, const Address& > m_rxTrace;
  
private:
  void ScheduleNextTx();
  void ScheduleStartEvent();
  void ScheduleStopEvent();
  void ConnectionSucceeded(Ptr<Socket>);
  void ConnectionFailed(Ptr<Socket>);
  void HandleRead (Ptr<Socket>);
};

/**
 * \ingroup traffic
 *
 * \brief Factory to create customized packets for the ns3::TrafficApplication.
 *
 * This base class defines an API for creating packets. The packet factory is
 * used by ns3::TrafficApplication and possibly other generic applications
 * which send opaque packets. Thus code for experiments can create subclasses
 * which generate special packet sequences with customized contents or tags.
 */
class TrafficPacketFactory : public Object
{
public:
  static TypeId GetTypeId();

  TrafficPacketFactory();

  /**
   * \return pointer to created Packet
   * 
   * Pure virtual class method for creating packet instances.
   */
  virtual Ptr<Packet> CreatePacket() = 0;
};

/**
 * \ingroup traffic
 *
 * \brief Packet factory which creates copies of a prototype packet.
 *
 * Whenever requested to create a packet this factory clones a prototype. The
 * prototype must be set before using the factory, otherwise it will create
 * NULL packets.
 */
class CloneTrafficPacketFactory : public TrafficPacketFactory
{
public:
  static TypeId GetTypeId();

  CloneTrafficPacketFactory(Ptr<Packet> prototype = NULL);

  /**
   * \brief Set packet to be cloned each time.
   */
  void SetPrototype(Ptr<Packet> prototype);

  /**
   * Creates a copy of the prototype and returns a pointer to it.
   *
   * \return pointer to the created clone
   */
  virtual Ptr<Packet> CreatePacket();

protected:
  /// Prototype packet
  Ptr<Packet>   m_prototype;
};

/**
 * \ingroup traffic
 *
 * \brief Packet factory which creates simple zeroed packets of a configured
 * size.
 *
 * Whenever requested to create a packet this factory creates a new empty
 * packet of a configured size.
 */
class SimpleTrafficPacketFactory : public TrafficPacketFactory
{
public:
  static TypeId GetTypeId();

  SimpleTrafficPacketFactory();

  /**
   * Creates a packet, applies  and returns a pointer to it.
   *
   * \return pointer to the created packet
   */
  virtual Ptr<Packet> CreatePacket();

  /**
   * Set size of packets generated. Useful for compatibility setters.
   */
  void SetSize(unsigned int size);

  /**
   * Return size of packets generated. Useful for compatibility getters.
   */
  unsigned int GetSize() const;

protected:
  /// Size of packets generated.
  unsigned int  m_size;
};

} // namespace ns3

#endif
