/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2006,2007 INRIA, 2009 Timo Bingmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as 
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Based on yans-wifi-channel.cc by 
 * Mathieu Lacage <mathieu.lacage@sophia.inria.fr>
 * Minor Modifications: Timo Bingmann <timo.bingmann@student.kit.edu>
 */
#include "ns3/packet.h"
#include "ns3/simulator.h"
#include "ns3/mobility-model.h"
#include "ns3/net-device.h"
#include "ns3/node.h"
#include "ns3/log.h"
#include "ns3/pointer.h"
#include "ns3/object-factory.h"
#include "ns2ext-wifi-channel.h"
#include "ns2ext-wifi-phy.h"
#include "propagation-loss-model.h"
#include "propagation-delay-model.h"

NS_LOG_COMPONENT_DEFINE ("Ns2ExtWifiChannel");

namespace ns3 {

TypeId 
Ns2ExtWifiChannel::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::Ns2ExtWifiChannel")
    .SetParent<WifiChannel> ()
    .AddConstructor<Ns2ExtWifiChannel> ()
    .AddAttribute ("PropagationLossModel", "A pointer to the propagation loss model attached to this channel.",
                   PointerValue (),
                   MakePointerAccessor (&Ns2ExtWifiChannel::m_loss),
                   MakePointerChecker<PropagationLossModel> ())
    .AddAttribute ("PropagationDelayModel", "A pointer to the propagation delay model attached to this channel.",
                   PointerValue (),
                   MakePointerAccessor (&Ns2ExtWifiChannel::m_delay),
                   MakePointerChecker<PropagationDelayModel> ())
    ;
  return tid;
}

Ns2ExtWifiChannel::Ns2ExtWifiChannel ()
{}
Ns2ExtWifiChannel::~Ns2ExtWifiChannel ()
{
  NS_LOG_FUNCTION_NOARGS ();
  m_phyList.clear ();
}

void 
Ns2ExtWifiChannel::SetPropagationLossModel (Ptr<PropagationLossModel> loss)
{
  m_loss = loss;
}
void 
Ns2ExtWifiChannel::SetPropagationDelayModel (Ptr<PropagationDelayModel> delay)
{
  m_delay = delay;
}

void 
Ns2ExtWifiChannel::Send (Ptr<Ns2ExtWifiPhy> sender, Ptr<const Packet> packet, Ptr<const WifiPhyTag> tag) const
{
  Ptr<MobilityModel> senderMobility = sender->GetMobility ()->GetObject<MobilityModel> ();
  NS_ASSERT (senderMobility != 0);
  uint32_t j = 0;
  for (PhyList::const_iterator i = m_phyList.begin (); i != m_phyList.end (); i++, j++)
    { 
      if (sender != (*i))
        {
          // For now don't account for inter channel interference
          if ((*i)->GetChannelNumber() != sender->GetChannelNumber())
              continue;
          
          Ptr<MobilityModel> receiverMobility = (*i)->GetMobility ()->GetObject<MobilityModel> ();
          Time delay = m_delay->GetDelay (senderMobility, receiverMobility);
          double rxPowerDbm = m_loss->CalcRxPower (tag->GetTxPowerDbm(), senderMobility, receiverMobility);
          NS_LOG_DEBUG ("propagation: txPower="<<tag->GetTxPowerDbm()<<"dbm, rxPower="<<rxPowerDbm<<"dbm, "<<
                        "distance="<<senderMobility->GetDistanceFrom (receiverMobility)<<"m, delay="<<delay);
          Ptr<Packet> packetcopy = packet->Copy ();
          Ptr<WifiPhyTag> tagcopy = Create<WifiPhyTag>(*tag);
          Ptr<Object> dstNetDevice = m_phyList[j]->GetDevice ();
          uint32_t dstNode;
          if (dstNetDevice == 0)
            {
              dstNode = 0xffffffff;
            }
          else
            {
              dstNode = dstNetDevice->GetObject<NetDevice> ()->GetNode ()->GetId ();
            }
          Simulator::ScheduleWithContext (dstNode,
                                          delay, &Ns2ExtWifiChannel::Receive, this, 
                                          j, packetcopy, tagcopy, rxPowerDbm);
        }
    }
}

void
Ns2ExtWifiChannel::Receive (uint32_t i, Ptr<Packet> packet, Ptr<WifiPhyTag> tag,
                            double rxPowerDbm) const
{
  m_phyList[i]->StartReceivePacket (packet, tag, rxPowerDbm);
}

uint32_t 
Ns2ExtWifiChannel::GetNDevices (void) const
{
  return m_phyList.size ();
}
Ptr<NetDevice> 
Ns2ExtWifiChannel::GetDevice (uint32_t i) const
{
  return m_phyList[i]->GetDevice ()->GetObject<NetDevice> ();
}

void 
Ns2ExtWifiChannel::Add (Ptr<Ns2ExtWifiPhy> phy)
{
  m_phyList.push_back (phy);
}

} // namespace ns3
